#!/usr/bin/python3

import os, time, asyncio, sys, logging, argparse, traceback, yaml,\
    pprint

from camagick.executor import PipeExecutor, PipeMaker, PipeValidator, \
    with_pipe, pipe_spec_from_args, find_processors, PipeHelpRequested, \
    make_pipe_type, pipe_type_from_node_dict

from camagick import _version

logger = logging.getLogger("camagick")

from camagick.probe import strip_str

yaml.SafeDumper.add_representer(strip_str, lambda d, s: d.represent_str(str(s)))

class CAspyApplication:
    def __init__(self, *executors, argv=None, env=None):

        self._argv = argv if argv is not None else sys.argv
        self._env = env if env is None else os.environ.copy()

        self._setup_logging()

        if len(executors) > 0:
            self.executors = executors
        else:
            self.executors = self._handle_args(self._argv[1:])


    def _handle_args(self, args):
        opts, pipe_args = self._parse_cmdline(args)
        self._maybe_help(opts)

        def _title(l, default=''):
            if l[1] == '':
                return l[0]
            else:
                return default
                
        def _body(l):
            if l[1] == '':
                return l[2:]
            else:
                return l

        def _underline(txt, char='='):
            return ''.join(([char]*len(txt)))+'\n'

        def _format_body(ilines):
            olines = []
            new_paragraph = None
            for l in ilines:

                if len(l.strip()) == 0:
                    new_paragraph = True
                    olines.append('')
                    continue
                
                if l[0] not in (' ', '\t') and \
                   l[-1]==':' and \
                   new_paragraph == True:
                    olines.append(f'## {l[:-1]}\n')
                    continue
 
                new_paragraph = False               
                olines.append(l)

            return '\n'.join(olines)
        
        try:
            spec = self._obtain_spec(opts, pipe_args)
            self._maybe_save_spec(opts, spec)
            pipe = self._pipe_from_spec(spec)
            return [PipeExecutor(pipe)]

        except PipeHelpRequested as ph:
            ntype = pipe_type_from_node_dict(ph.node)
            d1 = ntype.__doc__
            if hasattr(ntype, "__init__"):
                d2 = ntype.__init__.__doc__

            head = f'Using The "{ph.name}" {ph.kind.title()}-Processor'
            print(head)
            print(_underline(head, '='))
            for def_title,doc in (('General Information', ntype.__doc__),
                                  ('Initialization', ntype.__init__.__doc__),
                                  ('', ntype.__call__.__doc__)):
                if doc is None:
                    continue

                lns = doc.split('\n')
                if len(lns[0]) == 0:
                    del lns[0]
                t = _title(lns, default=def_title)
                print(t)
                print( _underline(t, '-') )
                print( _format_body((_body(lns))) )

            sys.exit(0)


    def _maybe_help(self, opts):
        if opts.list_all:
            self._print_flow()
            sys.exit(0)


    def _obtain_spec(self, opts, pipe_args):
        if opts.yaml:
            filename = '/dev/stdin' if opts.yaml == '-' else opts.yaml
            with open(filename, 'r') as f:
                raw_spec = yaml.safe_load(f)
        else:
            raw_spec = pipe_spec_from_args(pipe_args)

        return self._validate_spec(raw_spec)


    def _maybe_save_spec(self, opts, spec):
        if opts.save:
            print(yaml.safe_dump(spec))
            sys.exit(0)


    def _parse_cmdline(self, args):
        self._par = argparse.ArgumentParser(
            prog="caspy",
            usage="%(prog)s [options] <pipe>",
            description="Swiss Army Knife for scientific data collection.",
            epilog="Try `-list-all` for a list of processors."
        )

        self._par.add_argument('-version', action='version', version=_version.version,
                               help="prints version information and exits")
        self._par.add_argument('-list-all', action="store_true",
                               help="list available data flow and processing pipe modules")
        self._par.add_argument('-save', action="store",
                               help="don't execute the pipeline, just dump command line "
                               "arguments as YAML to stdout")
        self._par.add_argument('-yaml', action="store",
                               help="load processing pipeline from YAML file")

        _opts, _pargs = self._par.parse_known_args(args)
        
        return _opts, _pargs
    

    def _print_flow(self):

        mod_dict = {}
        max_len = 0

        print('Collecting information about available processors...')
        
        for ft in ('flow', 'source', 'sink', 'pipe'):
            proc_dict = find_processors(f'camagick.{ft}', 'Processor',
                                        noisy_missing=True)
            mod_dict[ft] = proc_dict
            max_len = max(max_len, *[len(i) for i in proc_dict])

        print()

        hints = {
            'source': ' (use as `--from-<name> ...`)',
            'sink':   ' (use as `--to-<name> ...`)',
            'pipe':   ' (use as `--<name> ...`)',
            'flow':   ' (mind the bracket syntax)'
        }        
        
        for ft, pipes in mod_dict.items():            
            
            print(f'{ft} processors{hints[ft]}:\n')
            for pname, pobj in pipes.items():

                docstr = pobj.__doc__ if pobj.__doc__ is not None else pobj.__init__.__doc__
                if docstr is not None:
                    parts = docstr.split("\n")
                    short = next(iter(filter(lambda x: len(x) > 0, parts)))
                else:
                    short = "(no documentation available)"

                fmt = '  {name:' + f'{max_len}' + '}  {short}'
                print(fmt.format(name=pname, short=short))
            print('\n')

        print('Try `--<processor> ?` for detailed processor-specific help.')

        
    def _setup_logging(self):
        logging.basicConfig()
        logger.setLevel(self._env.get("CASPY_LOG_LEVEL", "WARNING"))


    def _validate_spec(self, raw_spec):
        val = PipeValidator()
        return val.validate(raw_spec)


    def _pipe_from_spec(self, spec, validate=True):
        if validate:
            s = self._validate_spec(spec)
        else:
            s = spec
        pm = PipeMaker()
        with_pipe([s], pm)
        return pm.pipe
        
        
    async def run(self):
        await asyncio.gather(*[e.run() for e in \
                               filter(lambda x: x is not None,
                                      self.executors)],
                             return_exceptions=False)


def main(args=None, env=None):

    if args is None:
        args = sys.argv

    if env is None:
        env = os.environ.copy()

    app = CAspyApplication(argv=args, env=env)

    try:
        asyncio.run(app.run())
    except KeyboardInterrupt:
        logger.info(f'msg="Good bye."')


if __name__ == "__main__":
    main()



from camagick.processor import ProcessorBase
import numpy as np


def _parse_slice(s):
    parts = s.split(':')
    
    if len(parts)==1:
        return None if len(parts[0])==0 else int(parts[0])

    if len(parts) > 0:
        sparam = [int(p) if len(p)>0 else None for p in parts]
        return slice(*sparam)

    raise RuntimeError('Why are we here?')


class Processor(ProcessorBase):    
    '''
    Slices, a.k.a. selects, a subset of each dataset.
    '''

    def __init__(self, *rois):
        '''
        Iteratively selects regions of interests from data.

        Args:
            *rois: each unnamed parameter is either a `slice()`
              object, a string of the format "start:stop:step"
              or "index", or `None`. If it's a string, a slice
              object is constructed from the parsed string values.
              If it's `None`
        '''

        if len(rois)==1 and isinstance(rois[0], str):
            r = rois[0].split(',')
            rois = r

        # ROI tag. Will be used for naming the output data, defaults to "roi"
        #self.roi_tag = tag if tag is not None else "roi"

        # ROR params, one for each dimension.
        self.roi_params = []
        for r in rois:
            if type(r) in (type(None), int, type(slice(0))):
                self.roi_params.append(r)
            elif isinstance(r, str):
                self.roi_params.append(_parse_slice(r))
        
    
    async def __call__(self, data=None, context=None):
        ''' Executes the ROI selection. Context is ignored. '''
        try:
            return {
                key:val[*self.roi_params] for key,val in data.items()
            }
        except Exception as e:
            print("ROI:", self.roi_params)
            print("Data:", data)
            raise



from caproto.asyncio.server import Context as ServerContext
from caproto.server import PVGroup, pvproperty, PVSpec
from caproto import ChannelData
import caproto

from contextlib import suppress

import asyncio, logging

logger = logging.getLogger(__name__)


class DynamicMergeDict:
    ## El-Cheapo dictionary class that _dynamically_ returns items
    ## from a number of sub-dictionaries. Need this to manage the .pvdb
    ## of all common CAspy IOCs.
    
    def __init__(self):
        self.subs = []
        
    def add_sub(self, sub):
        self.subs.append(sub)

    def __getitem__(self, key):
        for s in self.subs:
            try:
                return s[key]
            except KeyError:
                pass
        raise KeyError(key)


    def __setitem__(self, key, val):
        raise RuntimeError('Nope.')


    def __repr__(self):
        return ' '.join([str(s) for s in self.subs])


    def __len__(self):
        return sum([len(s) for s in self.subs])


    def __contains__(self, key):
        for s in self.subs:
            if key in s:
                return True
        return False


class IocServer(object):
    ## Singleton IOC server class.
    ##
    ## Only this class ever spans an IOC. All other "IOCs" are actually
    ## just PV databases that eventually end up served by this one.
    ##
    ## Usage:  IocServer(pv_class)
    ##
    ## ...where pv_class has a .pvdb property that actually exposes PVs.
    
    def __new__(cls, *args, **kw):
        if not hasattr(cls, 'instance'):
            cls.instance = super(IocServer, cls).__new__(cls)
            cls.instance.__init_once__(*args, **kw)
            
        cls.instance.__init_always__(*args, **kw)
        return cls.instance

    
    def __init_once__(self, ioc):
        self.ioc_list = []
        self.refcnt = 0
        self.dyn = DynamicMergeDict()

    def __init_always__(self, ioc):
        self.ioc_list.append(ioc)
        self.dyn.add_sub(ioc.pvdb)


    @property
    def pvdb(self):
        return self.dyn


    async def startup(self):
        if self.refcnt == 0:
            self._name = f'IocServer'
            self.ctx = ServerContext(self.pvdb)
            self.task = asyncio.create_task(self.ctx.run(), name=self._name)
        self.refcnt += 1


    async def shutdown(self):
        if self.refcnt > 0:
            self.refcnt -= 1
        else:
            self.task.cancel()
            with suppress(asyncio.CancelledError):
                await self.task


class InstaOutputPVDB:
    ## Simple Output IOC class
    ##
    ##   .publish(name, data)
    ##
    ## ...makes sure that `data` ends up in the local EPICS
    ## network as a pv named `name`. PV is created on first
    ## .publish().
    
    def __init__(self):
        self.pv_dict = {}


    def _shape_size(self, s):
        if len(s)>1:
            return s[0] * self._shape_size(s[1:])
        elif len(s)==1:
            return s[0]
        else:
            return 1


    async def _publish_awaitable(self, name, val):
        # Returns an awaitable which sets val -> pv[name],
        # or creates pv[name] with a default value if it doesn't exist.

        print('writing PV:', name, '->', val)
        
        if name in self.pv_dict:
            if isinstance(val, str):
                return await self.pv_dict[name].write(val)
            else:
                return await self.pv_dict[name].write(val[()])

        else:
            #print('publish:', name, val)
            l = self._shape_size(val.shape)
            flat = val.flatten()
            dtype = type(flat[0].item())
            logger.info(f'msg="Creating PV" pvname="{name}" tag="{name}" length={l} '
                        f'shape={val.shape} dtype={dtype}')

            additional = dict(
                max_length=l
            )
            
            if dtype == str:
                additional.update(dict( max_length=40, ))
                dtype=caproto.ChannelType.STRING

            self.pv_dict[name] = PVSpec(name=name,
                                        value=flat,
                                        dtype=dtype,
                                        **additional).create(None)


    async def publish(self, name, val):
        await self._publish_awaitable(name, val)


    async def publish_many(self, set_dict, raise_errors=False):
        '''
        Publishes `data` as PV `name`.

        If the PV doesn't yet exist in `.pvdb`, a new PV is
        created.
        '''

        result = await asyncio.gather(*[
            self._publish_awaitable(name,val) for name,val in set_dict.items()
        ], return_exceptions=False)

        for r in result:
            if isinstance(r, Exception):
                logger.error(r)
                if raise_errors:
                    raise r
        

    
    @property
    def pvdb(self):
        return self.pv_dict



class InstaInputPVDB:
    '''
    Insta-PVDB specifically for the purpose of ingesting data.

    Exposes a .pvdb (to be used with whatever ServerContext contstruction),
    and exposes a .wait_for_input(), which releases once there's input
    available.
    '''

    def __init__(self, pv_map, hang=True,
                 suppress_initial=False,
                 suppress_same=False,
                 verify=None):
        '''
        Args:
            pv_map: map name -> value with the PVs to serve.

            suppress_initial: if set to `True`, the readout lock won't
              be lifted for the initial value -- only on explicit setting
              of the value.

            suppress_same: if set to `True`, the readout lock won't be
              lifted if the value written to PV is the same as the one
              the PV already has.
        
            verify: if not `None`, it's expected to be an async
              callable with the signature `verify(pvname, value)`
              called for incoming values, before processing them.
              If the value needs to be modified, the modified value
              needs to be returned. If `CaprotoValueError` is raised,
             then the value will not be updated and the CA protocol
             will propagate an error event instead.              
        '''

        self._hang = hang
        self._suppress_same = suppress_same
        self._suppress_initial = suppress_initial
        self._run_startup = True
        
        self._pvdb = {}
        self._pv_map = pv_map
        self._verify = verify

        self._input_lock = asyncio.Lock()

        for name,value in pv_map.items():

            if isinstance(value, str):
                self._pvdb[name] = PVSpec(
                    value=value,
                    name=name,
                    put=self._on_input,
                    dtype=caproto.ChannelType.STRING,
                    max_length=40,
                ).create(group=None)

            else:
                self._pvdb[name] = PVSpec(
                    value=value,
                    name=name,
                    put=self._on_input,
                ).create(group=None)

        for k in self._pvdb:
            logger.info(f'msg="Serving input PV" name={k}')


    async def _on_startup(self):
        # This should be done only once, on startup (but cannot be done
        # in __init__ since we need to await a lock acquisition.
        self._run_startup = False
        if (self._suppress_initial == True) and (not self._input_lock.locked()):
            print('initial suppress:', [k for k in self._pvdb.keys()])
            await self._input_lock.acquire()


    async def _on_input(self, pvinst, val):
        if self._verify is not None:
            v = await self._verify(pvinst.name, val)
        if self._hang and self._input_lock.locked():
            if not (self._suppress_same and val == pvinst.value):
                print('releasing:', pvinst.name, val, pvinst.value)
                self._input_lock.release()
            else:
                logger.info(f'msg="Suppressing unchanged PV value" '
                            f'pv="{pvinst.name}" '
                            f'value={val}')
                raise ValueError(f'value unchanged ({val})')
        if (self._verify is not None) and (v is not None):
            return v


    async def wait_for_input(self):
        # If we're already waiting for new input, the lock cannot be acquired --
        # so we need to wait.
        # Once we're free, we re-acquire the lock (to avoid repeated new_input()
        # until the next time the input condition is met).

        if (self._run_startup == True):
            await self._on_startup()

        if self._hang:
            print('waiting for:', [k for k in self._pvdb])
            await self._input_lock.acquire()


    @property
    def pvdb(self):
        return self._pvdb    



import pytest, asyncio, time, traceback
import multiprocessing as mp

from camagick.source.scratch import Processor as GenSource
from camagick.source.ioc import Processor as IocSource
from camagick.source.epics import Processor as EpicsSource

from camagick.sink.ioc import Processor as IocSink
from camagick.sink.epics import Processor as EpicsSink
from camagick.sink.summary import Processor as SummarySink

from camagick.sink.hdf5 import Processor as Hdf5Sink

from camagick.flow.chain import Processor as ChainFlow
from camagick.flow.fanout import Processor as FanoutFlow

from camagick.pipe.pace import Processor as Pace
from camagick.pipe.only import Processor as Only
from camagick.pipe.exclude import Processor as Exclude

from camagick.executor import PipeExecutor

import tempfile, os, random, h5py
import numpy as np

@pytest.fixture(scope='function')
def h5tmp():
    tmp = tempfile.NamedTemporaryFile(delete=False)
    tmp.close()
    yield tmp.name
    os.unlink(tmp.name)
    

async def test_h5_append(h5tmp):

    print(f"HDF5: {h5tmp}")

    h5_path = f'{h5tmp}#group'+'{gi}/{tag}{di}'

    sink = Hdf5Sink(h5_path, index='idx')

    nr_pts = 20
    scalar = np.random.random((nr_pts,))
    array = np.random.random((nr_pts, 23))

    for i in range(nr_pts):
        await sink(
            data={
                'scalar': scalar[i],
                'array':  array[i]
            },
            context={
                'gi': i % 2,
                'di': (int(i/10)+1)*10,
                'idx': int(i/2) % 5
            })

    with h5py.File(h5tmp, 'r') as h5:
        assert (h5['/group0/scalar10'][()] == scalar[0:10:2]).all()
        assert (h5['/group0/array10'][()]  ==  array[0:10:2]).all()
        assert (h5['/group1/scalar10'][()] == scalar[1:10:2]).all()
        assert (h5['/group1/array10'][()]  ==  array[1:10:2]).all()
        
        assert (h5['/group0/scalar20'][()] == scalar[10::2]).all()
        assert (h5['/group0/array20'][()]  ==  array[10::2]).all()
        assert (h5['/group1/scalar20'][()] == scalar[11::2]).all()
        assert (h5['/group1/array20'][()]  ==  array[11::2]).all()


async def test_h5_overwrite(h5tmp):
    print(f"HDF5: {h5tmp}")

    sink = Hdf5Sink(h5tmp, index='idx', mode='o')

    nr_pts = 20
    half_pts = int(nr_pts/2)
    array = np.array([[i+j*10 for i in range(6)] for j in range(nr_pts)])

    for i in range(half_pts):
        await sink(
            data={ 'array': array[i] },
            context={ 'idx': i })

    for i in range(half_pts):
        await sink(
            data={ 'array': array[i+half_pts] },
            context={ 'idx': i })
        

    with h5py.File(h5tmp, 'r') as h5:
        assert (h5['/array'][()] == array[half_pts:]).all()
    

    # since we didn't specify the top-up '+' flag,
    # writing beyond the end of the dataset should raise
    with pytest.raises(RuntimeError):
        await sink(data={'array': array[0]},
                   context={'idx': half_pts+2})


async def test_h5_reset(h5tmp):
    print(f"HDF5: {h5tmp}")

    sink = Hdf5Sink(h5tmp, index='idx', mode='x')

    nr_pts = 20
    half_pts = int(nr_pts/2)
    array = np.array([[i+j*10 for i in range(6)] for j in range(nr_pts)])

    for i in range(half_pts):
        await sink(
            data={ 'array': array[i] },
            context={ 'idx': i })

    for i in range(5):
        await sink(
            data={ 'array': array[half_pts+i] },
            context={ 'idx': i })
        

    with h5py.File(h5tmp, 'r') as h5:
        assert h5['/array'].shape[0] == 5
        assert (h5['/array'][()] == array[half_pts:half_pts+5]).all()
    

    # since we didn't specify the top-up '+' flag,
    # writing beyond the end of the dataset should raise
    with pytest.raises(RuntimeError):
        await sink(data={'array': array[0]},
                   context={'idx': half_pts+2})



from camagick.processor import SinkBase
import tifffile, logging, traceback, os

## Data sink to store (append) stuff in a HDF5 DataArray

logger = logging.getLogger(__name__)

class Processor(SinkBase):
    def __init__(self, pathfmt: str, start: int = 0, meta: list = None):
        '''
        Saves data into (a series of) TIFF files with dynamic naming.

        Each dataset generates one TIFF file per run. The name and
        path of the TIFF file are generated from scratch on every save,
        so that writing series of files, e.g. with increasing counters,
        is possible.
        
        Args:
            pathfmt: Path format for the tiff file. You can use
              the following variables:
                - `{seq}`: a sequential number (starting with 0 by default),
                  which is being incremented by 1 on every save -- cannot
                  be reset
                - `{suffix}`: the final component (the one after the last ':')
                  of the PV name
                - `{label}`: the key of the data item
                - any key of a data item available in the `context` section
                  of input data
            start: where to start the internal sequential counter

            meta: comma-seprated list of context tags to store in the TIFF
              as metadata. Note that this subject to whatever TiffFile supports.
              This is stored using the `metadata=...` argument to
              `tifffile.imwrite(...)`. So the metadata typically ends up as
              a JSON string on the first page of the Tiff frames object.
              (Load with `frames = tifffile.TiffFile(...)`, and view with
              `frames.pages[0].tags['ImageDescription'].value ...`)
        '''
        super().__init__()
        self.path_fmt = pathfmt
        self.seq = start
        self.meta_tags = meta if meta is not None else []


    def _collect_meta(self, context):
        try:
            meta = {}
            for k in self.meta_tags:
                # Need to convert the type from a numpy data type,
                # to a python type.
                v_raw = context[k]
                if hasattr(v_raw, "shape") and len(v_raw.shape)>0:
                    v_type = type(v_raw.flatten()[0].item())
                    meta[k] = [ v_type(k) for k in v_raw.flatten() ]
                elif hasattr(v_raw, "shape") and len(v_raw.shape)==0:
                    v_type = type(v_raw.flatten()[0].item())
                    meta[k] = v_type(v_raw)
                else:
                    meta[k] = v_raw
            return meta
        except KeyError as e:
            logger.error(f'msg="Data set required as metadata not found" key="{e}"')
            raise RuntimeError(f'Requested metadata {e} not found')


    def ensure_folder_of(self, path):
        d = os.path.dirname(path)
        if not os.path.exists(d):
            os.makedirs(d)
            return
        if not os.path.isdir(d):
            raise RuntimeError(f'msg="Path is required to be a folder folder" path="{p}"')
        

    async def __call__(self, data=None, context=None):

        if data is None:
            return

        format_params = dict()

        if context is not None:
            format_params.update(context)
        else:
            context = {}

        result = {}
        
        for k,v in data.items():
            path = None
            meta = {}
            try:
                suf = k.split(':')[-1]
                format_params.update(dict(seq=self.seq, suffix=suf, label=k))

                path = self.path_fmt.format(**format_params)

                tfopts = {}

                tfopts.update(dict(compression='zlib'))

                if len(self.meta_tags) > 0:
                    tfopts.update(dict(metadata=self._collect_meta(context)))
                    logger.info('msg="Have metadata" ' +
                                ' '.join([f'{m}={v}' for m,v in tfopts['metadata'].items()]))

                self.ensure_folder_of(path)

                tifffile.imwrite(path, v, **tfopts)
                result[k] = path
                
                #logger.info(f'msg="Writing TIFF" tag="{k}" path="{path}" shape="{v.shape}"')

            except Exception as e:
                traceback.print_exc()
                logger.error(f'msg="Error saving {k}" reason="{e}" shape="{v.shape}" '
                             f'path="{path}" '+
                             ' '.join([f'{ck}="{cv}"' for ck,cv in format_params.items()])+
                             ' '.join([f'{ck}="{cv}"' for ck,cv in meta.items()]))
        
        self.seq += 1

        return result



# Load data, attrs, and meta-information (shapes etc) from array storage
# (e.g. HDF5, Zarr, TIled, ...)

import logging, time, math, random, asyncio, importlib, functools

from camagick.processor import ProcessorBase, SourceBase
from camagick.probe import opcast
from numpy import random as np_rand
from numpy import array, zeros, ones

from camagick.storage.base import probe_kind, find_loader, DatasetUnavailable

logger = logging.getLogger(__name__)

def _split_at_brackets(res):
    # Extracts the part between brackets [...] if it exists
    if res[-1] == ']':
        i = res.rfind('[')
        if i >= 0:
            return res[0:i], res[i+1:-1]
        else:
            raise RuntimeError(f'Missing bracket for {res}')
    return res, None


def _make_slicer(s):
    # Generates a tuple of slice() objects from a string
    # of the form "a:b:c,d:e:f,...", where a,b,c ...
    # are integers. Returns "None" if not possible.
    slicers = []
    try:
        for spec in s.split(','):
            nr = spec.split(':')
            if len(nr) == 0:
                slicers.append(slice(None))
            elif len(nr) == 1:
                slicers.append(int(nr[0]) if len(nr[0])>0 else slice(None))
            elif len(nr) == 2:
                slicers.append(slice(int(nr[0]) if len(nr[0])>0 else 0,
                                     int(nr[1]) if len(nr[1])>0 else -1))
            elif len(nr) == 3:
                slicers.append(slice(int(nr[0]) if len(nr[0])>0 else 0,
                                     int(nr[1]) if len(nr[1])>0 else -1,
                                     int(nr[2]) if len(nr[2])>0 else 1))
            else:
                raise RuntimeError(f'{spec} is not a valid slicer')
        return tuple(slicers)
    except ValueError:
        return None


class Processor(SourceBase):

    def __init__(self,                 
                 url: str,
                 kind: str = 'auto',
                 wait=False,
                 **data):
        '''
        Access array data and metadata in storage.

        Some array formats (specifically Zarr) can be stored in various
        backends like S3, accessed via more complex URL formats (e.g.
        via s3:// or https://...). This processor makes no assumptions
        but simply passes on the parameter to the underlying storage
        backend.

        Args:
            url: Location of the data file or main array.
        
            kind: Which storage backend to use. If this is set to "auto",
              we try to guess which storage that would be. Typically this is
              done by parsing the ending of "location" and comparing it against
              a list of well-known endings ("hdf5", "h5", "zarr", "nx", "n5",
              ...), but as the list of supported formats expands, we tend to
              get more creative. It is advisable to explicitly specify the
              format if in doubt. Currently supported keys "zarr" and "hdf5"
              here.

            wait: if set to `True`, waits if data is unavailable
              (`FileNotFoundError` or `DatasetUnavailable`)

            **data: Data to query. The key here is the name inside `caspy`,
              the value is the path inside the array file. This can point to:
               - a typical dataset (e.g. `/group/subgroup/dataset`), in which case
                 we'll return the data array.
               - a slice of a dataset (e.g. `/group/subgroup/dataset[int:int:int]`
                 in which case we'll return the data of the slice
               - request for shape information (e.g. `/dataset.shape`), in which
                 case we'll return the shape
               - request for attributes of groups or datasets (`/dataset.attrs[key]`
                 or `/group.attrs[key]'), in which case we'll return the
                 corresponding attribute.
               - request for sub-items via a `.keys` suffix (the underlying Python
                 object should support the `.keys()` method)
        '''

        # We expect from all array storage systems to support the following:
        #  - slicing via dataset object dataset[...], dataset[()], ...
        #  - access to groups/subgroups/folders via obj['group/subgroup/dataset']
        #  - access to .attrs for groups and datasets.
        #  - access to dataset shapes via dataset.shape

        self._loader_url = url
        self._loader_kind = kind

        # Prepare the data retrieval. Store a map as:
        # { key: (path, action) }, where "key" is the data name inside caspy,
        # "path" is the object location within the data file, and "action"
        # is a callable that returns the desired data (slicing or attr key)
        self._keys = {}

        # waits if FileNotFoundError or DatasetUnavailable errors encountered
        self._wait = wait
        self._unavailable = None

        
        class _slicer_helper:
            # Helps with delaying slicer to a later point (the actual point
            # is to store the slicing information for later use)
            def __init__(self, *s, attr=None):
                self._sli = s if attr is None else s[0]
                self._attr = attr
            def __call__(self, data):
                if self._attr is None:
                    return data[*(self._sli)]
                else:
                    attr = getattr(data, self._attr)
                    if callable(attr):
                        return attr()[self._sli]
                    else:
                        return attr[self._sli]


        # Build up the request database (self._keys).
        # This is a { tag: (inside_path, reader_proc) } dictionary, where:
        #   - "tag" is the name within CAmagick (as usual)
        #   - "inside_path" is the path within the array storage of
        #     the object we want to read
        #   - "reader_proc" is a callable of the signature proc(obj)
        #     that will return the data, when passed the array object
        #     (i.e. loaded dataset or group object)
        for k,v in data.items():
            base, brk = _split_at_brackets(v)
            if brk is not None:
                # Brackets -- either a slicer, or an attrs key
                sli = _make_slicer(brk)
                if sli is not None:
                    # It's a data slicer.
                    action = _slicer_helper(*sli)
                elif base.endswith('.attrs'):
                    r = base.rfind('.')
                    suffix = base[r+1:]
                    # No slicer, just a string we'll use to get an attribute
                    base = base[:r]
                    action = _slicer_helper(brk, attr=suffix)
                else:
                    raise RuntimeError(f'Invalid slicer for {base}: {brk}')

            else:
                # No brackets -- we return either a shape, or the data
                if base.endswith('.shape'):
                    r = base.rfind('.')
                    suffix = base[r+1:]
                    base = base[:r]
                    action = lambda x: x.shape
                elif base.endswith('.keys'):
                    r = base.rfind('.')
                    suffix = base[r+1:]
                    base = base[:r]
                    action = lambda x: array([k for k in sorted(x.keys())])
                else:
                    action = lambda x: x[()]

            self._keys[k] = (base, action)


    def _get_loader(self, real_url):
        loader_kind = self._loader_kind
        return find_loader(kind=(loader_kind if loader_kind not in (None, 'auto') \
                                 else probe_kind(real_url)))


    def _get_url(self, **params):
        url_fmt = self._loader_url
        real_url = url_fmt.format(**params)
        if not hasattr(self, "_prev_url") or (real_url != self._prev_url):
            logger.info(f'msg="New input URL" url="{real_url}"')
            self._prev_url = real_url
        return real_url


    async def __call__(self, data=None, context=None):
        ctx = context if context is not None else {}
        result = {}
        url = self._get_url(**ctx)
        Loader = self._get_loader(url)
        reqd_elements = { k:v[0] for k,v in self._keys.items() }
        async with Loader(url, mode=None, **reqd_elements) as df:
            for k,v in self._keys.items():
                action = v[1]
                dobj = df[k]
                result[k] = action(dobj)
        return result




import caproto.asyncio.client as ca

import xarray, sys, argparse, logging, time, math, os, asyncio, traceback, pprint

from camagick.helpers import FpsCounter
from camagick.stash import ZarrStash
from camagick.reader import AsyncReader
from camagick.processor import ProcessorBase, SourceBase

from camagick.probe import opcast

logger = logging.getLogger(__name__)

class Processor(SourceBase):
    '''
    Reading from existing EPICS PVs

    The module retrieves data from EPICS process variables (PVs) that
    exists in your network. There are various tuning possibilities,
    including using other EPICS PVs as "guides" -- this means that the
    reading process will only be performed when the "guides" change
    to a certain value for the first time after having _not_ had that
    value.

    CLI Examples:
      Various readout scenarii, with data output to `--to-summary` for
      demonstration purposes:

      - Reding with auto-generated tag names:
        ```
        caspy --from-epics EXAMPLE:iffy --to-summary
        caspy --from-epics EXAMPLE:iffy EXAMPLE:jedda --to-summary
        ```

      - Reading multiple with explicitly named tags:
        ```
        caspy --from-epics my_scalar=EXAMPLE:iffy \\
                           my_array=EXAMPLE:jedda \\
              --to-summary
        ```

      - Reading when "guide-PV" blinks to 0, with an explicit dwell time:
        ```
        caspy --from-epics my_scalar=EXAMPLE:iffy \\
                           my_array=EXAMPLE:jedda \\
                           when=\\"EXAMPLE:mark==0\\" \\
                           dwell=0.2 \\
              --to-summary
        ```

    YAML Examples:
      Use these with `caspy -y ...`:

      ```
      chain:
      - from: "epics"
        params:
          my_scalar: "EXAMPLE:jedda"
          my_array: "EXAMPLE:iffy"
          when: "EXAMPLE:mark==0"
          dwell: 0.2
      - to: summary
      ```
    '''

    def __init__(
            self,
            *simple_pvs,
            mode: str = '',
            prefix: str = '',
            guide: str = '',
            when: str = '',
            dwell: float = 0.1,
            **tagged_pvs
    ):
        '''
        Initialization the EPICS-PV source

        Subscribes to all variables in `pvname` and returns their contents
        every time all the conditions in `guide` are met for the first time.

        Args:
          *simple_pvs: PVs to return as data. Each must be a string,
            and each string will be used as its own tag. As such, it must
            _not_ have any Python non-symbol characters (like ":" or ".").
            To still make it possible to build EPICS PV names in a simple way,
            the actual PV address will be built as `f{prefix}{}`, where
            `prefix` is the corresponding parameter.

          mode: determines when the `.__call__()` operator returns the data.
            It all boils down to two decisions necessary when acquiring more
            than one single PV: (1) when is a composite dataset considered
            ready for processing, and (2) which data goes into a dataset?
            Supported modes:

              - hasty: deliver data when /any/ new data is available,
                discard data that has been retrieved. This means that
                there will almost never be a point when all required
                data keys are available at once, unless there's a
                sufficiently large `dwell` time.
                This is the default when `guide` is empty.

              - sloppy: similarly to hasty, but don't discard data that
                has been retrieved. This means that except for the very
                first few retrievals, there will always be a full data
                set, but some data items may be old and others may be
                fresh. This is the behavior that's closest to polling data.

              - continuous: like sloppy, but don't block waiting for new
                data. Just return whatever is in the queue, and don't clean
                up the queue. This is similarly to what continously polling
                all the variables would do, except that no polling is actually
                performed. (ACHTUNG: IOCs that depend on polling to actually
                trigger a hardware readout will *not* therfore work as
                expected!)

              - thorough: return data only when /all/ data items are acquired,
                and discard data afterwards. Every successful query will
                have a full set of fresh, new data. But if one PV isn't
                updated regularly, data flow will stall.

              - guided: return data when the guide condition is fulfilled
                for the first time after having /not/ been fulfilled.
                This is the default when `guide` is non-empty.

              - clustered: save a timestamp for each incoming data point,
                then cluster together data points with similar timestamps.
                Return data when a cluser is "complete", i.e. when incoming
                data points clearly don't belong to the same cluster
                anymore. `dwell` time is ignored here.

            If you have a traditional move-acquire-process loop at the
            base of your experimental orchestration, then the "clustered"
            is likely to give the best results. Alternatively, you can use
            "guided", possibly with a suitable `dwell`-time, if you have a
            clear and reliable marker of when your device's acquisition
            process is finished (note that EPICS, by its TCP/IP nature,
            is notoriously unreliable as to the order and timing of packets,
            so it will really just boil down to your `dwell`).

            "sloppy" and "thorough" will give good results for continous
            data, or as a visualisation aid (i.e. where there's no harm
            when data points from different acquisitions may accidentally
            interleave).

            "hasty" may be useful with a fitting `dwell` time, otherwise
            you'll end up sending data through your processing pipeline
            one item at a time. This /may/ me acceptable if your pipeline
            is not dependent (or sensitive) to having a full data set
            to work with.

          prefix: Prefix to prepend to each of the `simple_pvs` PV names.

          guide: key-value pairs of EPICS variables and values to change
            to for the reading condition to be met. (FIXME: will change soon)

          when: same as `guide`

          dwell: Time to wait after guide condition is met until data
            is collected. This is necessary because EPICS is inherently a
            /non-realtime/ TCP-IP based protocol. This means that even if
            the sending IOC does everything right and signals the readiness
            (via guides) only /after/ the data is available, there's still
            no guarantee that the data will arrive in time at the consuming
            EPICS client.

         **tagged_pvs: PVs specified as `tag=EPICS:pv:name`, accessing
           an EPICS PV named "EPICS:pv:name", and making it accessible
           to the `caspy` pipeline under the tag "tag"
        '''

        if when is not None and len(when)>0:
            guide = when

        if len(mode) == 0:
            mode = 'sloppy' if len(guide)==0 else 'guided'
        if guide in (None, '') and mode == 'guided':
            raise RuntimeError(f'msg="Guided mode requested for EPICS reader, but no guide specified"')

        if guide not in (None, '') and (mode not in ('guided', '', None)):
            raise RuntimeError(f'msg="Guides specified for wrong mode" mode="{mode}"')

        self.options = {
            'guides': self._make_guides(guide),
            'pvs': tagged_pvs,
            'dwell': dwell,
            'mode': mode
        }

        for simple in simple_pvs:
            if ':' in simple:
                pvtag = simple.replace(':', '_').replace('.', '_')
            else:
                pvtag = simple
            old_pv = self.options['pvs'].get(pvtag, None)
            new_pv = f'{prefix}{simple}'
            if old_pv is not None:
                raise RuntimeError(f'PV tag "{pvtag}" intended for "{new_pv}" '
                                   f'already present for PV "{old_pv}"')
            self.options['pvs'][pvtag] = new_pv


        self._pv2tag = {}
        for tag,pv in self.options['pvs'].items():
            old = self._pv2tag.get(pv, None)
            if old is not None:
                e = f'tag={tag} new_pv={pv} old_pv={old} msg="Double-tag error"'
                logger.error(e)
                raise RuntimeError(e)
            self._pv2tag[pv] = tag

        super().__init__()


    def _make_guides(self, gspec):
        if gspec is None or len(gspec)==0:
            return None
    
        parts = gspec.split('==')
        
        if len(parts) == 2:
            return { parts[0]: opcast(parts[1]) }


    async def startup(self):
        await self.init_epics(**self.options)


    async def shutdown(self):
        await self.epics_reader.disconnect()


    async def init_epics(self, guides=None, pvs=None, dwell=None, mode=''):
        self.epics_reader = AsyncReader(*[ v for k,v in pvs.items() ],
                                        guide_dict=guides,
                                        auto_xarray=False,
                                        dwell=dwell,
                                        mode=mode)

        await self.epics_reader.connect()


    async def __call__(self, data=None, context=None):
        if data is None:
            data = {}

        try:
            incoming = await self.epics_reader.wait_for_incoming()
            my_data = { self._pv2tag[k]:v for k,v in incoming.items() }

            #my_data.update(data)
            return my_data

        except Exception as e:
            traceback.print_exc()
            logging.error("Data retrieve failed: %r", e)



from camagick.processor import ProcessorBase, SinkBase
from camagick.storage.hdf5 import ArrayAccess as H5ArrayAccess
from camagick.storage.base import RefuseIndex


import numpy as np
import h5py, logging, os, random, time

logger = logging.getLogger(__name__)

class Processor(SinkBase):
    '''
    Saves data as a dataset in a HDF5 file.

    The HDF5 file / dataset is opened explicitly for each dataset,
    on each run through the pipeline. Files that don't exist are
    created (though not filesystem folders), and files that do
    are appended to (`h5py` mode "a").

    The sink tries to make it easy to obtain uniform, synchronous,
    well-defined arrays of data of the same length. It does so
    mostly by filling up empty frame slots when there isn't any
    data available, or by overwriting data if necessary, according
    to the user specification. This is governed according to the
    `mode` initialization parameter, which determines what to
    do when a dataset already exists.
    '''
    
    def __init__(self,
                 path: str,
                 mode: str = 'a+',
                 index: str = '',
                 strict: bool = True,
                 timeout: float = 0.0,
                 ):
        '''
        Args:
            path: the path format. Generally, the expected syntax is
              "<HDF5-file>#<path-within-file>". The part beyond "#"
              represents the path to the dataset within the HDF5 file.
              The path within file may be just a simple dataset, or may
              contain HDF5 group elements ("#/path/to/group/dataset").
              All groups that don't exist are created.
        
              Note that this is a *format* which will be used for each
              incoming data set, so it should contain at least some
              differentiating element for substitution. Available elements
              are:
                - all the context keys
                - `index` for the designated numeric index of the
                  pipeline run (see `index`)
                - `tag` for the name / tag of the dataset being processed
              
              If the path within HDF5 ends in "/", or is an empty string,
              then it is assumed to be a group, and the format suffix `"{tag}"`
              is automatically assumed.

            mode:  Must be one of `a` (append), `o` (overwrite) or
              `x` (reset), possibly in combination with `+` (top-up):
        
                 - `a`: require that the data point be always appended
                   to the current dataset -- never overwrite or delete
                   existing data.
        
                 - `o`: overwrite already existing data points (frames).

                 - `O`: same as `o`, but additioally delete and re-create
                   whole dataset entries, if the shape or data type of the
                   existing dataset does not match the new request.

                 - `x`: reset the entire dataset (i.e. delete all existing
                   data points) if a data point already exists at the
                   specified index.

                 - `+`: top-up marker will fill up empty slots, but without
                    actually writing any data to them. They will remain
                    filled with default HDF5 settings according  to the data
                    type. If top-up is not specified, otherwise the operation
                    will fail if index is higher than the one of the next
                    slot in turn.
        
            index: name of the context variable which contains the current
              frame index (preferrably integer). If this is an empty string,
              no index is used and essentially every operation is an "append".

            strict: if this is `True` (the default), then any saving errors
              being raised will be allowed to bubble up to the pipe processor.
              Typically, this means the application is being shut down.
              When `False`, then `RefuseIndex` type of errors (typically when
              a frame index is out of place) are being reported, but ignored.

            timeout: If the target HDF5 file is locked, how long to wait for
              unlocking. Default is 5 seconds. Set to 0 to disable waiting,
              in which case a locked file will result in failing to save.
        '''
        super().__init__()
        
        parts = path.split('#')
        
        if len(parts) == 1:
            self.h5file = parts[0]
            self.h5grp = '/{tag}'
        elif len(parts) == 2:
            self.h5file, self.h5grp = parts
        else:
            raise RuntimeError(f'msg="Bad HDF5 path, expected <file>[#<group>[/<dataset>]]" '
                               f'path="path"')

        if (len(self.h5grp)==0) or \
           (self.h5grp[-1]=='/') or \
           self._has_group(self.h5file, self.h5grp):
            self.h5grp += (('/' if self.h5grp[-1] != '/' else '') + '{tag}')

        logger.info(f'msg="Initializing HDF5 saver" file="{self.h5file}" group="{self.h5grp}"')

        self._index_ctx_tag = index
        self._stacker_mode = mode
        self._current_file_and_node = (None, None)
        self._strict = strict
        self._unlock_timeout = timeout


    def _has_group(self, fpath, grp):
        try:
            with h5py.File(fpath, 'r') as h5:
                if isinstance(h5[grp], h5py.Group):
                    return True
        except KeyError: pass
        except OSError: pass
        return False


    def get_index(self, context):
        if self._index_ctx_tag not in (None, ''):
            try:
                if self._index_ctx_tag[0] == '@':
                    _index = context[self._index_ctx_tag[1:]]
                else:
                    _index = context[self._index_ctx_tag]
            except KeyError as e:
                avail = ','.join(context.keys())
                err = \
                    f'msg="Index key not available in this frame" ' \
                    f'expected="{self._index_ctx_tag}" ' \
                    f'available="{avail}"'
                raise RefuseIndex(err)
        else:
            _index = None

        return _index    
        

    def handle_index_error(self, err, tag=None):
        if self._strict or not isinstance(err, RefuseIndex):
            raise err
        logger.error(err)
        logger.error(f'msg="Ignoring frame index error as instructed" tag={tag}')


    async def __call__(self, data=None, context=None):
        
        if data is None:
            data = {}
        if context is None:
            context = {}

        try:
            _index = self.get_index(context)
        except RefuseIndex as e:
            self.handle_index_error(e)
            logger.warning(f'msg="Skipping frame altogether"')
            return
        
        async with H5ArrayAccess(
                store=self.h5file.format(**context), mode=self._stacker_mode,
                **{ k:self.h5grp.format(tag=k, **context) for k in data.keys()}) as stor:
            try:
                await stor.push_frame(_index, **data)
            except RefuseIndex as e:
                self.handle_index_error(e)



#!/usr/bin/python3

import os, sys, asyncio, time, logging
from camagick.probe import *

logger = logging.getLogger(__name__)

class Application:

    def __init__(self, args=None, env=None):
        self._args = args or sys.argv.copy()
        self._env  = env or os.environ.copy()

    async def startup(self):
        self._probe = make_probe(self._args[1:])
        logger.debug(f'input={self._args[1:]} probe={self._probe}')


    async def shutdown(self):
        pass
        

    async def run(self, period=0.1):
        if not hasattr(self, "_probe"):
            await self.startup()
            shut_me_down = True
        else:
            shut_me_down = False
            
        try:
            while self._probe():
                await asyncio.sleep(1)
            logger.debug(f'msg="Probe negative" expr={self._probe}')
        except Exception as e:
            from traceback import format_exc
            logger.error(f'msg="Probe error, traceback follows" detail={e}')
            logger.error(traceback.format_exc())
        finally:
            if shut_me_down:
                self.shutdown()


def main(args=None, env=None):
    app = Application(args or sys.argv, env or os.environ)
    asyncio.run(app.run())


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import asyncio, time, logging

from caproto.sync import client as ca_client
from caproto.asyncio.client import Context
from caproto import CaprotoTimeoutError
import numpy as np

import copy

from xarray import DataArray
from functools import partial
from contextlib import suppress

logger = logging.getLogger(__name__)

class PvRetry(RuntimeError):
    '''
    Raised by GuidedPvReader when the PVs are not yet ready / guide does
    not yet signal readiness for signal readout.
    '''
    def __init__(self, *p, **kw):
        super().__init__(*p, **kw)


class DataInbox:
    ## Base class to manage various types of readout strategies.
    ## All have in common that data is (possibly) kept in a cache
    ## until a specific readout condition is met.
    ##
    ## Waiting is implemented by means of an async lock.
    
    def __init__(self, dwell=0.2):
        # This is where we cache the data.
        # We use full PV names as keys.
        # The payload ("data hold") is:
        # {
        #    'tag': short,
        #    'data': [list of datapoints],
        #    'ts': [timestamps]
        # }
        self._incoming_data = {}
        self._dwell_time = dwell


    def _release_lock(self):
        with suppress(RuntimeError):
            self._readout_lock.release()


    async def _readout_lock_acquire(self, timeout=None):
        if timeout is None:
            await self._readout_lock.acquire()
        else:
            await asyncio.wait_for(self._readout_lock.acquire(), timeout)


    def _have_complete_set(self):
        ## Checkes whether a complete data set (as expected) is available.
        have = [ k for k in self._incoming_data.keys() ]
        for x in self._expected:
            if x not in have:
                return False # don't have a full set, NO READOUT
        return True


    def _trim_backlog(self, num_max=1):
        ## Removes all but `num_max` data entries from every PV's backlog.
        ## This MUST BE CALLED if they don't reset the backlog altogether.
        for pvname,hold in self._incoming_data.items():
            if len(hold['data']) > num_max:
                del hold['data'][:num_max]
                del hold['ts'][:num_max]


    def _reset_backlog(self):
        self._incoming_data = {}


    async def create(self, expected):
        '''
        Called when within the designated asyncio loop.

        Expected to perform initialization steps that can only be
        performed from within the loop.

        Args:
            expected: List of expected full PV set. Used for various decision-making
              procedures (see also ._have_complete_set()).
        '''
        self._expected = expected
        self._readout_lock = asyncio.Lock()


    def ingest(self, pvname, tag, response):
        '''
        Called when incoming data is arriving.

        Args:
            pvname: full PV name
            tag: short name (not used? may be "none"?)
            response: CAproto response object; contains `.data`
        '''
        hold = self._incoming_data.setdefault(
            pvname,
            {
                'tag': tag,
                'data': [],
                'ts': []
            }
        )
        hold['data'].append(response)
        hold['ts'].append(time.time())


    async def readout(self, timeout=None):
        '''
        Returns data, if this is the case according to readout policy.

        Might block until readout is due.
        '''
        if self._dwell_time > 0.0:
            await asyncio.sleep(self._dwell_time)

        # return only the last item of each key (not the whole incoming stack)
        return {
            k:{'data': v['data'][-1], 'ts': v['ts'][-1]} \
            for k,v in self._incoming_data.items()
        }


class ContinuousDataInbox(DataInbox):
    '''
    Continuous and uncoditional EPICS readout.
    No waiting for (new) data, no resetting of the readout
    buffer.
    '''
    def ingest(self, pvname, tag, response):
        super().ingest(pvname, tag, response)
        self._trim_backlog()
    

class LockingDataInbox(DataInbox):
    ## Base class that locks the ._readout_lock generally (at the
    ## beginning, and usually after readout), and only unlocks it
    ## when a new dataset is available.
    ## Essentially everything except ContinuousDataInbox depends
    ## on this.

    def can_readout(self):
        ## This is checked on every ingest. Default behavior for
        ## most modes is "yes, we can read out". But some specific
        ## modes -- i.e. "thorough" and possibly "guided" -- will
        ## overwrite this to implement a more sophisticated method.
        return True

    async def create(self, expected):
        await super().create(expected)
        await self._readout_lock_acquire()

    def ingest(self, pvname, tag, response):
        super().ingest(pvname, tag, response)
        self._trim_backlog()
        if self.can_readout():
            self._release_lock()

    async def readout(self, timeout=None):
        await self._readout_lock_acquire(timeout)
        d = await super().readout()
        return d


class GuidedDataInbox(LockingDataInbox):

    def __init__(self, guide_dict, **kwargs):
        super().__init__(**kwargs)

        if guide_dict is None:
            guide_dict = {}
        
        # This is a dictionary guide-pv <-> test lambda
        self.guides = {}
        self.guides.update({
            k:v if hasattr(v, "__call__") else lambda k,x: x == v \
            for k,v in guide_dict.items()
        })

        # map EPICS name -> current True/False  evaluation of the guide signal.
        # Note that this is _not_ the guide trigger evaluation, i.e. the condition
        # of whether the waiting for this guide is finished and we're ready to
        # return data! For the latter to be fulfilled, the corresponding
        # guide_eval needs to be changing from 'False' to 'True'!
        self.guide_evals = { k:None for k in self.guides }


    async def create(self, expected):
        await super().create(expected)
        self.guide_trigger = {
            k:False for k in self.guides.keys()
        }


    def _guide_changed(self, pv_name, response):
        '''
        Called when a guide value changes. Checks if all guide conditions
        are met, and possibly calls the incoming hooks.
        '''

        d = response.data if len(response.data) > 1 else response.data[0]
        eval_result = self.guides[pv_name](pv_name,d)

        logger.info(f'msg="Guide update" pv={pv_name} value="{d}" type={type(d)}')

        if eval_result:
            if not self.guide_evals[pv_name]:
                # on eval switch False -> True: trigger!
                self.guide_trigger[pv_name] = True
        else:
            # eval False always kills the trigger
            self.guide_trigger[pv_name] = False
            
        self.guide_evals[pv_name] = eval_result

        #print(eval_result, self.guide_evals, self.guide_trigger, self.guides[pv_name](pv_name,d))



    def can_readout(self):

        # check whether we can do a data readout (all triggers must be True)
        guides_ok = all(self.guide_trigger.values())

        if not guides_ok:
            return

        return True


    def _invalidate_guide(self):
        for k in self.guide_trigger:
            self.guide_trigger[k] = False


    def ingest(self, pvname, tag, response):
        if (pvname in self.guides):
            #logger.info(f'msg="Guide changed" name={pv_name} value={d} '
            #            f'condition_ok={eval_result} '
            #            f'activated_ok={self.guide_evals[pv_name]} '
            #            f'all_ok={guides_ok}')
            self._guide_changed(pvname, response)
        return super().ingest(pvname, tag, response)


    async def readout(self, timeout=None):
        data = await super().readout()
        self._invalidate_guide()
        not_guide = lambda x: x[0] not in self.guides
        return { k:v for k,v in filter(not_guide, data.items()) }


class SloppyDataInbox(LockingDataInbox):
    '''
    Unlocks the readout as soon as _any_ data is available.
    Don't reset the readout (meaning that subsequent ingest
    operations will eventually return a full set of data,
    but possibly from different points in time).
    '''
    pass

        
class HastyDataInbox(SloppyDataInbox):
    '''
    Same as SloppyDataInbox (i.e. returns data on _any_ ingest),
    but reset the backlog on retrieve. This means that there
    will almost never be a full dataset available (unless 'dwell'
    is sufficiently large).
    '''
    async def readout(self, timeout=None):
        d = await super().readout(timeout)
        self._reset_backlog()
        return d


class ThoroughDataInbox(SloppyDataInbox):
    '''
    Returns the data only when a full dataset is available.
    Resets the backlog afterwards, so that a _new_ full dataset
    is required for new readout.
    '''

    def can_readout(self):
        return self._have_complete_set()


class ClusteredDataInbox(DataInbox):
    # See also: https://web.stanford.edu/class/cs345a/slides/12-clustering.pdf
    pass

    

class AsyncReader:
    '''
    Observes a "guide" variable to determine when a specific EPICS PV signal is
    available, then collects the PV signal (which can come in a list of other PVs).    

    To get as efficient as possible, this implementation actually
    uses subscriptions to the guide variables, and reacts on
    those. Calling .value() will wait asynchronously for the guide
    conditions to be fulflled, and this will be fairly efficient.

    But the most efficient way will be to subscribe a data
    processing callback using .subscribe(), which will be called
    only when all guide conditions are fulfilled and all data
    is available.
    '''

    def __init__(self,
                 *pvs: list,
                 guide_dict: dict = None,
                 mode: str = '',                 
                 auto_xarray: bool = True,
                 dwell: float = None):
        '''
        Initialises the reader.

        Args:
        
            pvs: The PVs to read out. (If not specified here, they can
              be specified later -- FIXME: really?)
              All PVs to be monitored are being subscribed to, and the newest
              version of incoming data is stored without backlog (!), until
              the guide conditions are all met.
        
            guide_dict: A dicitonary of guide variable(s) as keys, and a
              corresponding value to use, OR a callable of the signature
              `proc(key, value)`.
              The "guide condition" is "activated" for each individual guide
              when the callable returns `True` for the first time; the "guide
              condition" is "met" when _all_ guides have been activated.
                      
              When the guide condition is met, the current values of the PVs
              specified in `pvs` are returned as readout, and a new readout
              is only possible after a new reactivation. Note that for a new
              reactivation, a *de*activation of the individuals is necessary
              first, and that a *de*activation is *not* done automatically.
              Rather, it is expected that the guides deactivate themselves
              by taking on a value that doesn't meet the activation condition.

              Thus, guides are required to "blink" at least once for every
              data readout.
        
            mode: Readout mode, one of "hasty", "sloppy", "continuous",
              "thorough", "guided" or "clustered". See documentation of
              `camagick.source.epics` for a detailed explantion on the
              readout mode.

            guide_probe: A single probe object to signalize the guide condition.

            dwell: Time (in seconds) to wait after guide condition is met
              until data is collected. ACHTUNG: the wait is performed only
              for `.wait_for_data()`, not when using data callbacks!
        '''

        self.pvs = [v for v in pvs]
        
        for v in (guide_dict if guide_dict is not None else {}):
            if v not in pvs:
                self.pvs.append(v)
                continue
            logger.warning(f'msg="PV explicitly requested as data _and_ guide will '
                           f'not be returned as data" name={v}')
        
        # DEACTIVATED (old legacy code, not sure it's useful anymore...)
        #
        #    auto_xarray: If set to `True`, data is transformed to `xarray` data
        #      set rather than `numpy` arrays, and some implicit conversions and
        #      enhancements are triggered. These are... (FIXME: explain!)        
        self.auto_xarray = False
        #if auto_xarray:
        #    for k in self.pvs:
        #        if k.endswith("_SIGNAL"):
        #            kbase = k.replace('_SIGNAL', '')
        #            self.pvs.append(f'{kbase}_OFFSET')
        #            self.pvs.append(f'{kbase}_DELTA')


        if mode is None or len(mode) == 0:
            if guide_dict is None or len(guide_dict)==0:
                mode = 'sloppy'
            else:
                mode = 'guided'

            if dwell is None:
                dwell = 0.1
    
            logger.info(f'msg="EPICS readout mode autodetection" '
                        f'mode="{mode}" dwell={dwell}')
        else:
            logger.info(f'msg="Explicit EPICS readout mode" mode="{mode}"')

        # inbox needs to be initialized in a running asyncio loop.
        self.inbox = {
            'hasty': partial(HastyDataInbox, dwell=dwell),
            'sloppy': partial(SloppyDataInbox, dwell=dwell),
            'continuous': partial(ContinuousDataInbox, dwell=dwell),
            'thorough': partial(ThoroughDataInbox, dwell=dwell),
            'guided': partial(GuidedDataInbox, guide_dict=guide_dict, dwell=dwell),
            'clustered': partial(ClusteredDataInbox, dwell=0.0)
        }[mode]()

        self.ctx = None #ctx

        self.data_pvs = None
        self.incoming_hooks = []

        self._ref_cnt = 0


    async def __aenter__(self):
        if self._ref_cnt == 0:
            await self.connect()
        self._ref_cnt += 1
        return self


    async def __aexit__(self, *a, **kw):
        if self._ref_cnt > 0:
            self._ref_cnt -= 1
        if self._ref_cnt == 0:
            await self.disconnect()


    async def connect(self, ctx=None):
        
        if ctx is not None:
            self.ctx = ctx
            
        if self.ctx is None:
            logger.info(f'msg="Creating CA client context"')
            self.ctx = Context()
            self.own_ctx = True
        else:
            logger.info(f'msg="Reusing CA client context" ctx="{self.ctx}"')

        # map PV name -> Context PV, for all non-guide variables
        self.data_pvs = await self.ctx.get_pvs(*[v for v in self.pvs])

        await self.inbox.create(expected = [ d.name for d in self.data_pvs])
            
        #  map EPICS name -> Subscription obj, for all guides
        self.data_subscriptions = {}

        for d in self.data_pvs: # + self.guide_pvs:
            logger.info(f'Subscribing to data: {d.name}')
            self.data_subscriptions[d.name] = d.subscribe()
            self.data_subscriptions[d.name].add_callback(self._data_changed)
        

        # map EPICS name -> trigger-condition-fulfilled, for all guides.
        # When all of these are True, that's when we need to read
        # all the non-guide data.
        # Each of these becomes True when the guide changes value _to_
        # whatever was specified. It switches to False again when a data
        # readout is performed, or when it swithces value away _from_
        # whatever was specified.
        #
        # Note that this is subtly different from self.guide_evals!
        # While guide_evals is a True/False value per guide, representing
        # the current condition evaluation, the guide_trigger truth value
        # also indicates the *changing* to the True value.
        #
        # Also important to note: keeping this accurate for _all_ guides
        # at once is difficult when more than one guide is involved (values
        # will change with a slight delay). Hence the policy: once a guide
        # is considered triggered, it _stays_ triggered for as long as
        # its value doesn't chance, or data isn't read.

        logger.info(f"New CA context state: {self.ctx}")


    async def disconnect(self):
        if hasattr(self, "own_ctx") and self.own_ctx == True:
            await self.ctx.disconnect()
        

    def extract_data(self, response, pvName=None, others=None):
        '''
        Extracts "useful" data out of a response telegram.
        '''

        if response is None:
            return None

        if others is None:
            others = {}

        # Channel types can be: CHAR, DOUBLE, FLOAT, STRING, ENUM, LONG, INT.
        # The intention is to get an automatic useful native Python data type,
        # scalar or array. This means different things for different data
        # types.
        # In addition, we implement some heuristics to decorate waveforms
        # (== arrays) if our obscure array markers are present (shape, dimensions,
        # axis scaling -- to be documented ;-) )
        
        if response.data_type in (ca_client.ChannelType.STRING,):
            return response.data[0].decode('utf-8')
        
        elif response.data_type in (ca_client.ChannelType.DOUBLE,
                                    ca_client.ChannelType.FLOAT,
                                    ca_client.ChannelType.LONG,
                                    ca_client.ChannelType.INT,
                                    ca_client.ChannelType.ENUM):
            
            if len(response.data) == 1:
                return response.data[0]

            if not pvName or not pvName.endswith("_SIGNAL"):
                return response.data
            
            # If we have an array and it ends on _SIGNAL, we also try to
            # load _OFFSET and _DELTA for intrinsic scaling information
            o_name = pvName.replace("_SIGNAL", "_OFFSET")
            d_name = pvName.replace("_SIGNAL", "_DELTA")

            if o_name in others:
                offs = self.extract_data(others.get(o_name, 0))
            else:
                offs = 0

            if d_name in others:
                dlta = self.extract_data(others.get(d_name, 1))
            else:
                dlta = 1

            try:
                axis = offs+np.array(range(len(response.data)))*dlta
            except TypeError as e:
                # This happens when not all the data (e.g. `dlta` or `offs`
                # has arrived yet.
                axis = np.array([np.nan] * len(response.data))

            if self.auto_xarray:
                return DataArray(data=response.data, dims=["x"], coords=[axis])
            else:
                return response.data

        elif response.data_type in (ca_client.ChannelType.CHAR,):
            # This is a string -- return as such (ACHTUNG, this will break
            # "everything is a numpy array" data philosophy -- but we _want_
            # this. There's no other way to discern a uint8 from a string
            # later on.
            p = response.data
            s = bytes(p).decode('utf-8')
            return s

        elif response.data_type in (ca_client.ChannelType.STRING,):
            # This is a string -- return as such (ACHTUNG, this will break
            # "everything is a numpy array" data philosophy -- but we _want_
            # this. There's no other way to discern a uint8 from a string
            # later on.
            #print('have', repsonse.data)
            return response.data
        

        # else: how to handle ENUM / CHAR?
            
        else:
            logger.warning ("Unhandled data type: %r" % (response.data_type,))
            return response.data[0]


    def subscribe_incoming(self, proc):
        '''
        Registers a hook for processing incoming data.
        
        The hook will receive a dictionary with full PV
        name(s) as keys, and data as values, when guide condition is met.
        '''
        self.incoming_hooks.append(proc)


    def _data_changed(self, sub, response):
        '''
        Called when new data arrives.
        '''
        self.inbox.ingest(sub.pv.name, None, response)

    
    async def _get_incoming(self):
        '''
        Returns the currently incoming data (from subscriptions).
        We're using this instead of directly reading out the `._incoming_data`
        dictionary because, optinally, we're doing some mangling
        in `.extract_data()` that may or may not interfere with
        the data keys.
        '''

        # need to work on a copy here because _incoming_data might change
        # while we're waiting for new incoming data.
        tmp = await self.inbox.readout()

        return {
            k:self.extract_data(v['data'], k, others=tmp) \
            for k,v in tmp.items() \
            if (v is not None)
        }


    async def wait_for_incoming(self):
        '''
        Waits for incoming data. Returns only once per read-cycle
        (i.e. "marks" itself as called, so that repeated calls to
        this function sleep until new data is actually available,
        according to the ._readout_mode strategy).
        '''

        d = await self._get_incoming()
        return d


# Legacy name
GuidedAsyncReader = AsyncReader



from camagick.processor import ProcessorBase, SinkBase

import multiprocessing as mp
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import numpy as np

from camagick.executor import QuitApplication

import time, math, logging, concurrent, asyncio, pickle

logger = logging.getLogger(__name__)


def resample_array(data, num_bins):
    slices = np.linspace(0, data.shape[0], num_bins+1, True).astype(int)
    counts = np.diff(slices)
    mean = np.add.reduceat(data, slices[:-1]) / counts
    return mean


class MatplotlibDisplay:
    '''
    Stupidly simple class that uses Matplotlib to visualize numpy arrays.
    '''
    def __init__(self, rows=None, cols=None, flip_lr=True, transpose=False, sample_max=2056):
        self.flip_lr = flip_lr
        self.transpose = transpose
        self._rows = rows
        self._cols = cols
        self._sample_max = sample_max
        self.figure = plt.figure()
        #self.figure.show(True)
        self.panels = {}
        

    def _init_rows_cols(self, panel_names, rows, cols):
        self.panelNames = panel_names

        if rows is not None:
            self._rows = rows

        if cols is not None:
            self._cols = cols

        rows = self._rows if self._rows is not None else \
            (int(math.ceil(math.sqrt(len(self.panelNames)))))

        if rows == 0:
            logger.error(f'msg="ACHTUNG, no data"')
            rows += 1
        
        cols = self._cols if cols is not None else \
            int(math.ceil(len(self.panelNames)/rows))

        return rows, cols


    def init_display(self, *panel_names, rows=None, cols=None):
        rows, cols = self._init_rows_cols(panel_names, rows, cols)
        logger.info(f'geometry={rows}x{cols} displays={self.panelNames}')
        self.axes = { k: self.figure.add_subplot(rows, cols, i+1) \
                      for i,k in enumerate(self.panelNames) }
        self.last_update = {k:time.time() for k in self.panelNames }


    def _plot_norm_2d(self, img):
        mi = img.min()
        ma = img.max()
        return mi, ma, LogNorm(vmin=mi if mi > 0 else 1e-5,
                               vmax=ma if ma > 0 and ma > mi else 1)
    
        
        
    def _plot_2d(self, name, data, *kw):
        ax = self.axes[name]
        ax.set_title(name)

        if self.flip_lr:
            img = data[:,::-1]
        else:
            img = data

        if self.transpose:
            img = np.transpose(img)
        
        self._2d_norm = self._plot_norm_2d(img)

        return ax.imshow(img, norm=self._2d_norm[2], *kw)


    def _plot_1d(self, name, data, *kw):
        if len(data) > self._sample_max:
            d = resample_array(data, self._sample_max)
            logger.info(f'tag={name} msg="Data too large, rebinning" '
                        f'full_shape={data.shape} '
                        f'bin_shape={d.shape}')
        else:
            d = data
        ax = self.axes[name]
        ax.set_title(name)
        return ax.plot(d, *kw)


    def _update_1d(self, name, data, **kw):
        if len(data) > self._sample_max:
            d = resample_array(data, self._sample_max)
        else:
            d = data
        x = np.array(range(len(d)))
        
        self.panels[name][0].set_data((x, d))
        
        ax = self.axes[name]
        ylim = ax.get_ylim()
        
        #print(ylim)
        mi, ma = d.min(), d.max()
        
        if (ylim[0]<mi) or (ylim[1]>ma):
            ax.set_ylim(mi-(ma-mi)*0.05, ma+(ma-mi)*0.05)


    def _update_2d(self, name, data, **kw):
        if self.flip_lr:
            img = data[:,::-1]
        else:
            img = data

        if self.transpose:
            img = np.transpose(img)

        self.panels[name].set_data(img)

        mi, ma, norm = self._plot_norm_2d(img)
        if mi < self._2d_norm[0] or ma > self._2d_norm[1]:
            logger.info(f'msg="Updating norm" tag="{name}" vmin="{mi}", vmax="{ma}"')
            self.panels[name].set_norm(norm)
            self._2d_norm = (mi, ma, norm)


    def _panel_for_data(self, name, data):
        '''
        Returns the panel for data set `name`
        '''
        dims = len(data.shape)
        plot_func = getattr(self, f"_plot_{dims}d")
        plot_kwargs = {
            1: {},
            2: {}
        }[dims]

        if data is None:
            data = np.ndarray([2]*dims)

        pan = self.panels.get(name)
        if pan is None:
            pan = self.panels.setdefault(name, plot_func(name, data, **plot_kwargs))

        return plt
        

    @property
    def is_initialized(self):
        return hasattr(self, "axes")
        

    def update(self, panel, data=None):
        
        if data is None:
            return
        
        try:
            dims = len(data.shape)
            ax = self._panel_for_data(panel, data)
        except KeyError:
            logging.error("%s: no such display panel" % panel)
            return

        getattr(self, f"_update_{dims}d")(panel, data)
        self.figure.canvas.draw_idle()


    def run(self):
        plt.show()
        #pass


class ProcessPlotter:
    def terminate(self):
        plt.close('all')

    def callback(self):
        data = {}
        m0 = time.time()
        while self.pipe.poll():
            data = self.pipe.recv()
            if data is None:
                self.terminate()
                return False

        if (not self.display.is_initialized):
            if len(data)>0:
                self.display.init_display(*[k for k in data.keys()])

        #print('plot', [k for k in data.keys()])
        for name,val in data.items():            
            self.display.update(name, val)
            
        return True


    def __call__(self, pipe, period=0.03):
        self.pipe = pipe
        self.display = MatplotlibDisplay(transpose=True, flip_lr=False)
        intr = int(period*1000)
        logger.info(f'msg="Data refresh, milliseconds" interval={intr}')
        
        timer = self.display.figure.canvas.new_timer(interval=intr)
        timer.add_callback(self.callback)
        timer.start()
        
        self.display.run()


class Processor(SinkBase):
    '''
    Plots data as lines (1D arrays) or images (2D arrays).
    '''
    
    def __init__(self, only: str = ''):
        self._only = [ k for k in filter(lambda x: len(x)>0, only.split(',')) ]


    async def startup(self):
        self.my_pipe, other_pipe = mp.Pipe()
        self.plotter = ProcessPlotter()
        self.plotter_process = mp.Process(
            target=self.plotter, args=(other_pipe,), daemon=True
        )
        self.plotter_process.start()        


    async def shutdown(self):
        self.my_pipe.send(None)

    
    async def __call__(self, data=None, context=None):
        try:
            if self._only is not None and len(self._only) > 0:
                d = { k:v for k,v in \
                      filter(lambda x: x[0] in self._only, data.items()) }
            else:
                d = data

            with concurrent.futures.ProcessPoolExecutor() as pool:
                await asyncio.get_running_loop()\
                             .run_in_executor(pool, self.my_pipe.send, d)


            return { k:data[k] for k in \
                     filter(lambda x: x not in d, data.keys()) }
            
        except BrokenPipeError:
            logger.info(f'msg="Display closed by user"')
            raise QuitApplication(f'msg="Display closed by user"')



#!/usr/bin/python3

import os, time, asyncio, glob, sys, logging, argparse, traceback

import camagick.source

#camagick_sources = {
#    k:getattr(camagick.source, k) for k in \
#    filter(lambda x: hasattr(getattr(camagick.source, x), 'populate_argparser'),
#           camagick.source.__dir__())
#}

from camagick.source.epics import Processor as EpicsSource
from camagick.sink.summary import Processor as SummarySink
from camagick.sink.hdfdata import Processor as Hdf5Sink
from camagick.sink.tiffseries import Processor as TiffSink
from camagick.sink.epics import Processor as EpicsSink
from camagick.pipe.slice import Processor as SlicePipe
from camagick.pipe.npstat1 import Processor as Stat1Pipe
from camagick.pipe.reshape import Processor as ReshapePipe
from camagick.pipe.fanout  import Processor as FanoutPipe
from camagick.pipe.chain  import Processor as ChainPipe

logger = logging.getLogger("camagick")

def find_subapps(module):
    names = [
        k for k in filter(
            lambda x: hasattr(
                getattr(module, x),
                "populate_argparser"
            ),
            module.__dir__())
    ]
    
    return { n:getattr(module, n) for n in names }


def make_argparser(name):
    ## Parses main application command options
    cap = argparse.ArgumentParser(prog=name)

    #sources_sub = cap.add_subparsers(dest="sources")

    for s_name,s_mod in find_subapps(camagick.source).items():
        logger.info(f'source={s_name} module={s_mod}')
        cap.add_argument(s_name, action='append')
        s_mod.populate_argparser(cap)

    return cap


class CAMagickApplication:
    
    def __init__(self, pipeline):
        self.pipeline = pipeline

        
    async def run(self):

        await self.pipeline.startup()
        
        try:
            in_data = None

            while True:
                out_data = await self.pipeline(in_data)
                if (out_data is not None) and len(out_data)>0:
                    logger.warning(f'msg="Residual data keys" keys="{out_data.keys()}"')


        except Exception as e:
            logger.error(e)
            logger.error('Good bye')

            traceback.print_exc(e)

        finally:
            await self.pipeline.shutdown()


def main(args=None, env=None):

    if args is None:
        args = sys.argv

    if env is None:
        env = os.environ.copy()

    logging.basicConfig()
    logger.setLevel("INFO")

    #ap = make_argparser("camagick")
    #ap.print_help()

    #opt = ap.parse_args(args[1:])

    chain = ChainPipe(
        EpicsSource(
            pvprefix="KE:threshold_1:",
            pvname=['image'],
            guide={ 'KE:acquire': 1, },
            context={
                'width': 'asize1',
                'heigh': 'asize2'
            }
        ),
        
        ReshapePipe(
            shape=(1, 512, 1028)
        ),
        
        FanoutPipe(
            #SummarySink(),
            #Hdf5Sink("/tmp/foo.h5#mega"),
            TiffSink("/tmp/Xcurrent-{tag}-{seq}-layer.tif"),
            ChainPipe(
                #SlicePipe(),
                FanoutPipe(
                    Stat1Pipe("max"),
                    Stat1Pipe("sum")
                ),
                
                # ACHTUNG, using this without a prefix will shadow any
                # EPICS PVs by the same name as the incoming data -- it only
                # works like this if the previous processing replaces all
                # the input keys. Otherwise you need to specify a different
                # prefix.
                EpicsSink()
            )
            
        )
    )

    app = CAMagickApplication(chain)

    asyncio.run(app.run())

if __name__ == "__main__":
    main()
